/*

Copyright (C) 2003 Olivier VERLINDEN
    Service de Mecanique rationnelle, Dynamique et Vibrations
    Faculte Polytechnique de Mons
    31, Bd Dolez, 7000 MONS (Belgium)
    Olivier.Verlinden@fpms.ac.be

This file is part of EasyDyn

EasyDyn is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the
Free Software Foundation; either version 2, or (at your option) any
later version.

EasyDyn is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with EasyDyn; see the file COPYING.  If not, write to the Free
Software Foundation, 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
	
#include <string.h>
#include <math.h>
#include <fstream>
#include <iostream>
#include <iomanip>
//#include <time.h>

using namespace std;

#include <gsl/gsl_math.h>
#include <gsl/gsl_errno.h>
#include <gsl/gsl_linalg.h>
#include <gsl/gsl_blas.h>

#include <EasyDyn/sim.h>

gsl_vector_view fview,qview,qdview,qddview;
gsl_vector *qsto=0,*qdsto=0,*qddsto=0,*qddsto2=0,*qddcor=0;
gsl_permutation *perm;
gsl_matrix *jac=0;
ofstream DbgFile;

//---------------------------------------------------------------------------

void EasyDynsimError(char *texte)
{
 cout << "EasyDynsim ERROR: " << texte << "\n";
 exit(1);
}

//---------------------------------------------------------------------------

void InitEasyDynsim()

{
if (nbrdof==0) 
  EasyDynsimError("nbrdof must be initialized before calling InitFpmsSim");
if (application==0) 
  EasyDynsimError("application must be initialized before calling InitFpmsSim");
f=new double[nbrdof];
q=new double[nbrdof];
qd=new double[nbrdof];
qdd=new double[nbrdof];
fview=gsl_vector_view_array(f,nbrdof);
qview=gsl_vector_view_array(q,nbrdof);
qdview=gsl_vector_view_array(qd,nbrdof);
qddview=gsl_vector_view_array(qdd,nbrdof);
qsto=gsl_vector_alloc(nbrdof);
qdsto=gsl_vector_alloc(nbrdof);
qddsto=gsl_vector_alloc(nbrdof);
qddsto2=gsl_vector_alloc(nbrdof);
qddcor=gsl_vector_alloc(nbrdof);
perm = gsl_permutation_alloc(nbrdof);
jac=gsl_matrix_calloc(nbrdof,nbrdof);

int idof;
for (idof=0; idof<nbrdof; idof++)
   {
     q[idof]=0.0; qd[idof]=0.0; qdd[idof]=0.0;
   } 
}

//---------------------------------------------------------------------------

void EndEasyDynsim()

{
delete f; delete q; delete qd; delete qdd;
gsl_vector_free(qsto);
gsl_vector_free(qdsto);
gsl_vector_free(qddsto);
gsl_vector_free(qddsto2);
gsl_vector_free(qddcor);
gsl_permutation_free(perm);
gsl_matrix_free(jac);
}

//---------------------------------------------------------------------------

void WriteStateVariablesHeader(ostream &OutFile)

{
int idof;
OutFile << " time ";
for (idof=0;idof<nbrdof; idof++) 
  OutFile << " q" << idof << " qd" << idof << " qdd" << idof;
OutFile << " ";
}

//--------------------------------------------------------------------

void SaveStateVariables(ostream &OutFile)

{
int idof;
OutFile << " " << t;
for (idof=0;idof<nbrdof; idof++) 
  OutFile << " " << q[idof] << " " << qd[idof] << " " << qdd[idof];
OutFile << " ";
}

//--------------------------------------------------------------------

void ComputeResidual2(int *doflocked)

{
int idof;
ComputeResidual();
if (doflocked) for (idof=0; idof<nbrdof; idof++) 
   if (doflocked[idof]) f[idof]=0.0;
}

//--------------------------------------------------------------------

void Calcule_Matrice_J(double alphaM, double alphaC, double alphaK, double tol,
                       int *doflocked)

{
  int iddl,jddl;
  double *fsto;
  fsto=new double[nbrdof];
  ComputeResidual2(doflocked);
  for (iddl=0; iddl<nbrdof; iddl++) fsto[iddl]=f[iddl];
  for (jddl=0; jddl<nbrdof; jddl++)
    {
    q[jddl]+=alphaK*tol;
    qd[jddl]+=alphaC*tol;
    qdd[jddl]+=alphaM*tol;
    if (doflocked) { if (!doflocked[jddl]) ComputeResidual2(doflocked); }
    else ComputeResidual2(doflocked);
    for (iddl=0; iddl<nbrdof; iddl++)
      gsl_matrix_set(jac,iddl,jddl,(f[iddl]-fsto[iddl])/tol);
    q[jddl]-=alphaK*tol;
    qd[jddl]-=alphaC*tol;
    qdd[jddl]-=alphaM*tol;
    }
  for (iddl=0; iddl<nbrdof; iddl++) f[iddl]=fsto[iddl];
  delete fsto;
  if (doflocked) for (iddl=0; iddl<nbrdof; iddl++) if (doflocked[iddl])
    {
    for (jddl=0; jddl<nbrdof; jddl++)
      {
      gsl_matrix_set(jac,iddl,jddl,0);
      gsl_matrix_set(jac,jddl,iddl,0);
      }
    gsl_matrix_set(jac,iddl,iddl,1);
    }
}

//---------------------------------------------------------------------------

void StaticEquilibrium(int *doflocked)

{
  int nstep,s;
  double err;
  nstep=0;
  cout << "Equilibrium iterations: ";
  do
    {
    nstep++;
    cout << "+";
    if ((nstep%10)==1)
      {
	Calcule_Matrice_J(100,0,1,1E-5,doflocked);
	gsl_linalg_LU_decomp(jac,perm,&s);
      }
    ComputeResidual2(doflocked);
    gsl_linalg_LU_solve(jac,perm,&fview.vector,qddcor);
    gsl_vector_sub(&qview.vector,qddcor);
    err=gsl_blas_dnrm2(qddcor);
    if (DEBUG) cout << "Err=" << err << "\n";
    }
  while ((nstep<1000) && (err>(1E-8*sqrt(1.0*nbrdof))));
  if(nstep==1000) EasyDynsimError("No convergence after 1000 iterations");
  cout << " Bingo" << "\n";
}

//---------------------------------------------------------------------------

int NewmarkOneStep(double h, double &errqd, int *doflocked)
  {
  // Let's control the user assiduity
  if ((f==0) || (q==0) || (qd==0) || (qdd==0) || (qsto==0) || (qdsto==0) 
      || (qddsto==0) || (qddsto2==0) || (qddcor==0) || (jac==0)) 
    EasyDynsimError("NewmarkOneStep: Newmark variables not allocated");
  int iddl,nstep,s;
  double err,Beta=0.25,Gamma=0.5;
  // Copie de reference pour les accelerations pour calculer l'erreur
  gsl_vector_memcpy(qddsto2,&qddview.vector);
  // Prediction
  t+=h;
  for (iddl=0; iddl<nbrdof; iddl++)
      {
      q[iddl]+=h*qd[iddl]+0.5*h*h*qdd[iddl];
      qd[iddl]+=h*qdd[iddl];
      }
  if (DEBUG) cout << "Newmark iterations: ";
  // Correction
  nstep=0;	  
  do
    {
    nstep++;
    if (DEBUG) cout << "+";
    if ((nstep%4)==1)
      {
	Calcule_Matrice_J(1,Gamma*h,Beta*h*h,1E-2,doflocked);
        gsl_linalg_LU_decomp(jac,perm,&s);
      }
    ComputeResidual2(doflocked);
    //cout << "Residu" << "\n";
    //gsl_vector_fprintf(stdout,&residu.vector,"%g");
    gsl_linalg_LU_solve(jac,perm,&fview.vector,qddcor);
    //err=gsl_blas_dnrm2(qddcor)/sqrt(1.0*nbrdof);
    err=gsl_blas_dnrm2(qddcor)/(sqrt(1.0*nbrdof)*(1+gsl_blas_dnrm2(&qddview.vector)));
    gsl_vector_sub(&qddview.vector,qddcor);
    gsl_vector_scale(qddcor,Gamma*h);
    gsl_vector_sub(&qdview.vector,qddcor);
    gsl_vector_scale(qddcor,Beta*h/Gamma);
    gsl_vector_sub(&qview.vector,qddcor);
    //cout << "qd" << "\n";
    //gsl_vector_fprintf(stdout,&qdcopy.vector,"%g");
    //err=0;
    //cout << "Err=" << err << "\n";
    }
  while ((nstep<10) && (err>1E-8));
  if(nstep==10) return 1;
  if (DEBUG) cout << " Bingo" << "\n";
  // Calcul de l'erreur a partir de la variation du vecteur acceleration
  gsl_vector_sub(qddsto2,&qddview.vector);
  errqd=(h/12)*gsl_blas_dnrm2(qddsto2)/sqrt(1.0*nbrdof);
  if (!gsl_finite(errqd)) return 2;
  return 0;
}

//---------------------------------------------------------------------------

void NewmarkInterval(double tfinal, double &h, double hmax, int *doflocked)
// Procede a l'integration jusque tfinal

{
  // Let's control the user assiduity
  if ((f==0) || (q==0) || (qd==0) || (qdd==0) || (qsto==0) || (qdsto==0) 
      || (qddsto==0) || (qddsto2==0) || (qddcor==0) || (jac==0)) 
    EasyDynsimError("NewmarkInterval: Probably InitMbsSim was not called");
  int iddl, istep, code;
  double errqd,errqdmax=1E-6,timesto;
  // Time step verification
  if (h>hmax) h=hmax;
  // Copie de reference pour les accelerations
  while (t<tfinal) 
    {
    if ((t+1.4*h)>=tfinal) h=tfinal-t;
    timesto=t;
    gsl_vector_memcpy(qsto,&qview.vector);
    gsl_vector_memcpy(qdsto,&qdview.vector);
    gsl_vector_memcpy(qddsto,&qddview.vector);
    code=NewmarkOneStep(h,errqd,doflocked);
    if (DEBUG) cout << "time:" << t << " h=" << h
                    << " Erreur sur les positions: " << errqd << "\n";
    if ((code) || (errqd>errqdmax))  // la solution n'est pas convaincante
      {
        if (DEBUG)
          {
          if (code==1) cout << "\nNo convergence - ";
          if (code==2) cout << "\nNumerical trouble - ";
          if (gsl_finite(errqd)) if (errqd>errqdmax) 
			  cout << "\nLack of accuracy - ";
          }
        h*=0.5;
        if (DEBUG) cout << "Reducing time step to " << h << "\n";
        t=timesto;
        gsl_vector_memcpy(&qview.vector,qsto);
        gsl_vector_memcpy(&qdview.vector,qdsto);
        gsl_vector_memcpy(&qddview.vector,qddsto);
        if (h<1E-8) 
          EasyDynsimError("\nNo convergence with the smallest time step - STOP");
      }
    else 
      {
      // Let's save the information
      if (DEBUG) DbgFile << t << " " << h << " " << errqd << "\n";
      // The time step is increased if the error is really small
      if (errqd<0.25*errqdmax)
        {
        if (DEBUG) cout << "\nMmh.. Good ... Increasing time step\n";
        h*=2;
        if (h>hmax) h=hmax;
        }
      }
    }
}

//---------------------------------------------------------------------------

void NewmarkIntegration(double tfinal, double hsave, double hmax, 
			int *doflocked)
// Procede a l'integration complete sur des intervalles successifs 
// correspondant au pas de sauvegarde

{
  // Let's control the user assiduity
  if ((f==0) || (q==0) || (qd==0) || (qdd==0) || (qsto==0) || (qdsto==0) 
      || (qddsto==0) || (qddsto2==0) || (qddcor==0) || (jac==0)) 
    EasyDynsimError("NewmarkIntegration: Probably InitMbsSim was not called");
  // Ouverture du fichier de resultats et ecriture de la ligne d'entete
  char *nom;
  nom=new char[strlen(application)+5];
  strcpy(nom,application);
  strcat(nom,".res");
  cout << "Opening file " << nom << "\n" << "\n";
  ofstream ResFile(nom);
  if (!ResFile) EasyDynsimError("Cannot open file");
  WriteDataHeader(ResFile);
  // Ouverture du fichier DEBUG
  if (DEBUG)
      {
      strcpy(nom,application);
      strcat(nom,".dbg");
      cout << "Opening file " << nom << "\n" << "\n";
      DbgFile.open(nom);
      if (!DbgFile) EasyDynsimError("Cannot open file");
      }
  // Calcul des accelerations initiales et sauvegarde
  t=0;
  double h=0,dummy;
  NewmarkOneStep(0,dummy,doflocked);
  cout << "Sauvegarde\n";
  SaveData(ResFile);
  // Integration proprement dite
  int ipas=0;
  h=hmax;
  while (t<tfinal)
    {
      ipas++;
      cout << "Process up to time  " << hsave*ipas << " s" ;
      NewmarkInterval(hsave*ipas,h,hmax,doflocked);
      cout << " - Finished - Saving "; 
      SaveData(ResFile);
      cout << " - Done" << "\n";
    }
  // Fermeture des fichiers
  ResFile.close();
  if (DEBUG) DbgFile.close();
}

//---------------------------------------------------------------------------

extern "C" void dggev_ (char *JOBVL, char*JOBVR,
                        int *N, double *A, int *LDA,
                        double *B, int *LDB,
                        double *ALPHAR, double *ALPHAI, double *BETA,
                        double *VL, int *LDVL, double *VR, int *LDVR,
                        double *WORK, int *LWORK,
                        int *INFO);

// Variables globales du module de linearisation, liees a la resolution
// du probleme aux valeurs propres avec la routine DGGEV de LAPACK

void ComputePoles(int *doflocked, double freqmin, double freqmax)
// Procede au calcul des poles pour la configuration actuelle

{
  int NN,LWORK,INFO,*Liste;
  gsl_matrix *AA,*BB,*VR,*VL;
  double *ALPHAI, *ALPHAR, *BETA, *WORK;
  NN=2*nbrdof;
  // Allocation memory
  AA=gsl_matrix_alloc(NN,NN);
  BB=gsl_matrix_alloc(NN,NN);
  VR=gsl_matrix_alloc(NN,NN);
  VL=gsl_matrix_alloc(NN,NN);
  ALPHAR=new double[NN];
  ALPHAI=new double[NN];
  BETA=new double [NN];
  LWORK=12*NN;
  WORK=new double [LWORK];
  Liste=new int[NN];

  // Building eigenvalue problem matrices
  int iddl,jddl,ipole,jpole;
  // Mise a zero des matrices
  gsl_matrix_set_zero(AA);
  gsl_matrix_set_zero(BB);
  // Ajout de la contribution de la matrice masse
  // Attention, on transpose les termes dans AA et BB 
  // car LAPACK est ecrit en FORTRAN
  cout << "Constructing Stiffness matrix ...";
  Calcule_Matrice_J(0,0,1,1E-5,doflocked);
  cout << " Done\n";
  for (iddl=0; iddl<nbrdof; iddl++) for (jddl=0;jddl<nbrdof; jddl++)
     gsl_matrix_set(AA,nbrdof+jddl,nbrdof+iddl,-gsl_matrix_get(jac,iddl,jddl));
  cout << "Constructing Damping matrix ...";
  Calcule_Matrice_J(0,1,0,1E-3,doflocked);
  cout << " Done\n";
  for (iddl=0; iddl<nbrdof; iddl++) for (jddl=0;jddl<nbrdof; jddl++)
     gsl_matrix_set(BB,nbrdof+jddl,nbrdof+iddl,gsl_matrix_get(jac,iddl,jddl));
  cout << "Constructing Mass matrix ...";
  Calcule_Matrice_J(1,0,0,1,doflocked);
  cout << " Done\n";
  for (iddl=0; iddl<nbrdof; iddl++) for (jddl=0;jddl<nbrdof; jddl++)
     {
     gsl_matrix_set(AA,jddl,iddl,gsl_matrix_get(jac,iddl,jddl));
     gsl_matrix_set(BB,nbrdof+jddl,iddl,gsl_matrix_get(jac,iddl,jddl));
     gsl_matrix_set(BB,jddl,nbrdof+iddl,gsl_matrix_get(jac,iddl,jddl));
     }

  // Calling LAPACK routine and placing eigenvalues in ALPHAR and ALPHAI
  dggev_ ("N", "V", 
          &NN, AA->data, (int *) &(AA->tda), BB->data, (int *) &(BB->tda),
          ALPHAR, ALPHAI, BETA,
          VL->data, (int *) &(VL->tda), VR->data, (int *) &(VR->tda),
          WORK, &LWORK,
          &INFO);
  if (INFO<0) EasyDynsimError("Bad argument passed to DGGEV routine of LAPACK");
  if (INFO>0) EasyDynsimError("QZ Algorithm failed");
  for (ipole=0; ipole<NN; ipole++) 
    {
    if (fabs(BETA[ipole])<1E-15)
      { 
      BETA[ipole]=1.0; 
      ALPHAR[ipole]=1E30;
      ALPHAI[ipole]=1E30;
      }
    ALPHAR[ipole]/=BETA[ipole];
    ALPHAI[ipole]/=BETA[ipole];
    }
  // Building ordered list of poles
  for (ipole=0; ipole<NN; ipole++) Liste[ipole]=ipole;
  for (ipole=0; ipole<NN-1; ipole++) for (jpole=ipole; jpole<NN; jpole++)
    {
    if (ALPHAI[Liste[jpole]]<ALPHAI[Liste[ipole]])
           { 
           int itmp=Liste[ipole];
           Liste[ipole]=Liste[jpole]; 
           Liste[jpole]=itmp;
           }
    if (fabs(ALPHAI[Liste[jpole]]-ALPHAI[Liste[ipole]])<1E-15)
      if (ALPHAR[Liste[jpole]]<ALPHAR[Liste[ipole]])
           { 
           int itmp=Liste[ipole];
           Liste[ipole]=Liste[jpole]; 
           Liste[jpole]=itmp;
           }
    }
  if (DEBUG) for (ipole=0; ipole<NN; ipole++) if (ALPHAI[Liste[ipole]]>-1E-15)
    cout << "Pole " << Liste[ipole] << "=" << ALPHAR[Liste[ipole]]
             << "+-i" << ALPHAI[Liste[ipole]] << "\n";
  // Building complete results file
  int numpole, FREQMINMAX=0;
  double polenorm,freq,amor,coeff;
  char *nom;
  if ((freqmin!=0) || (freqmax!=1E30)) FREQMINMAX=1;
	
  nom=new char[strlen(application)+5];
  strcpy(nom,application);
  strcat(nom,".lst");
  ofstream lstfile(nom);
  if (!lstfile)
    {
    cout<<"\n Fichier "<<nom<<" impossible a ouvrir";
    exit(1);
    }
  lstfile << "  POLE       ALPHA         OMEGA        " 
        << "FREQ(Hz)   DAMP_RATIO \n";
	
  ofstream ls2file;
  if (FREQMINMAX)
    {
    strcpy(nom,application);
    strcat(nom,".ls2");
    ls2file.open(nom);
    if (!ls2file)
      {
      cout<<"\n Fichier "<<nom<<" impossible a ouvrir";
      exit(1);
      }
    }
  strcpy(nom,application);
  strcat(nom,".mod");
  ofstream modfile(nom);
  if (!modfile)
    {
    cout<<"\n Fichier "<<nom<<" impossible a ouvrir";
    exit(1);
    }
	
  for(ipole=0;ipole<NN;ipole++) if (ALPHAI[Liste[ipole]]>-1E-15) 
    {
    numpole=Liste[ipole];
    polenorm=hypot(ALPHAR[numpole],ALPHAI[numpole]);
    freq=ALPHAI[numpole]/(2*3.14159265);
    if (polenorm<1E-15) polenorm=1;
    amor=-ALPHAR[numpole]/polenorm;
    lstfile << setw(5) << ipole+1
            << " " << setw(13) << ALPHAR[numpole]
            << " " << setw(13) << ALPHAI[numpole]
            << " " << setw(13) << freq
            << " " << setw(13) << amor <<"\n";
    if ((freq>=freqmin) && (freq<=freqmax))
      {
      if (FREQMINMAX) ls2file << setw(5) << ipole+1
              << " " << setw(13) << ALPHAR[numpole]
              << " " << setw(13) << ALPHAI[numpole]
              << " " << setw(13) << freq
              << " " << setw(13) << amor <<"\n";
      modfile << setw(5) << ipole+1
              << " " << setw(13) << ALPHAR[numpole]
              << " " << setw(13) << ALPHAI[numpole]
              << " " << setw(13) << freq
              << " " << setw(13) << amor <<"\n";
      // Sauvegarde du mode
      if (ALPHAI[numpole]>1E-15) // cas d'un pole avec partie imaginaire
        {
        for (iddl=0; iddl<nbrdof; iddl++)
          modfile << gsl_matrix_get(VR,numpole,iddl) << " "
                  << gsl_matrix_get(VR,numpole+1,iddl) << endl;
        }
      else for (iddl=0; iddl<nbrdof; iddl++)  // cas d'un pole reel pur
        modfile << gsl_matrix_get(VR,numpole,iddl) << " 0.0" << endl;
      }
    }
  lstfile.close();
  if (FREQMINMAX) ls2file.close();
  modfile.close();
  
  // Cleaning memeory
  gsl_matrix_free(AA);
  gsl_matrix_free(BB);
  gsl_matrix_free(VR);
  gsl_matrix_free(VL);
  delete ALPHAR;
  delete ALPHAI;
  delete BETA;
  delete WORK;
  delete Liste;
  }

//---------------------------------------------------------------------------
